(* During inference, we will need an environment to keep track of
 * the user defined types *)

module LabelId = struct
  type t = Datatype.label_id
  let compare = Stdlib.compare
end

module LabelIdMap =
  Map.Make(LabelId)

type 't label2descr = 't Datatype.label_descr LabelIdMap.t

module TyConstrId = struct
  type t = Datatype.tyconstr_id
  let compare = Stdlib.compare
end

module TyConstrIdMap =
  Map.Make(TyConstrId)

type ('b,'t) datatype2descr = ('b,'t) Datatype.datatype_descr TyConstrIdMap.t

(* Environment for user defined data types *)
type ('b,'t) t = {
    datatypes : ('b,'t) datatype2descr ;
    labels : 't label2descr ;
  }

exception UnexpectedRecord

let empty = {
    datatypes = TyConstrIdMap.empty;
    labels = LabelIdMap.empty;
  }

let add_data_type m tdescr =
  TyConstrIdMap.add Datatype.(tdescr.name) tdescr m

let add_labels m ldescrs =
  List.fold_left (fun m ldescr ->
      LabelIdMap.add Datatype.(ldescr.label_name) ldescr m
    ) m ldescrs

let add_typedecl e tdescr = {
    datatypes = add_data_type e.datatypes tdescr;
    labels = add_labels e.labels Datatype.(tdescr.labels_descr);
  }

let descr_from_label { labels ; _ } (Datatype.Label l) =
  LabelIdMap.find (Datatype.Label l) labels

let type_descr_from_type { datatypes ; _ } (Datatype.Type tid) =
  TyConstrIdMap.find (Datatype.Type tid) datatypes

let label_index (Datatype.({ labels_descr ; _ })) l =
  let combine i ldescr = (i, Datatype.(ldescr.label_name)) in
  List.mapi combine labels_descr
  |> List.find (fun (_, l') -> l = l')
  |> fst

let map (type b1 b2 t1 t2)
  (f : (b1, t1) Datatype.datatype_descr -> (b2, t2) Datatype.datatype_descr)
  (env : (b1, t1) t) : (b2, t2) t
=
  TyConstrIdMap.to_seq env.datatypes
  |> Seq.map snd
  |> Seq.map f
  |> Seq.fold_left add_typedecl empty

(* [Temporary] Turns off warning 32 (unused value) *)
let () = ignore (empty, add_typedecl, descr_from_label,
                      type_descr_from_type, label_index)
