(* This is the target calculus of the sample client. It is a core System F. *)

(* -------------------------------------------------------------------------- *)

(* Types. *)

(* We include recursive types [FTyMu] in the target calculus, not only because
   we might wish to support them, but also because even if we disallow them,
   they can still arise during unification (the occurs check is run late), so
   we must be able to display them as part of a type error message. *)

(* Nominal or de Bruijn representation of type variables and binders. The
   nominal representation is more convenient when translating from ML to F,
   while the de Bruijn representation is more convenient when type-checking
   F. *)

type ('a, 'b) typ =
  | TyVar of 'a
  | TyArrow of ('a, 'b) typ * ('a, 'b) typ
  | TyProduct of ('a, 'b) typ list
  | TySum of ('a,'b) typ list
  | TyForall of 'b * ('a, 'b) typ
  | TyMu of 'b * ('a, 'b) typ
  | TyConstr of Datatype.tyconstr_id * ('a,'b) typ list

type tyvar = int

type nominal_type =
    (tyvar, tyvar) typ

type debruijn_type =
    (DeBruijn.index, unit) typ

type nominal_datatype_env =
    (tyvar, nominal_type) Env.t

type debruijn_datatype_env =
    (unit, debruijn_type) Env.t

(* -------------------------------------------------------------------------- *)

(* Terms. *)

(* Nominal representation of term variables and binders. *)

type tevar = string

type ('a, 'b) term =
  | Var of tevar
  | Abs of tevar * ('a, 'b) typ * ('a, 'b) term
  | App of ('a, 'b) term * ('a, 'b) term
  | Let of tevar * ('a, 'b) term * ('a, 'b) term
  | TyAbs of 'b * ('a, 'b) term
  | TyApp of ('a, 'b) term * ('a, 'b) typ
  | Tuple of ('a, 'b) term list
  | Proj of int * ('a, 'b) term
  | Inj of (('a,'b) typ list) * int * ('a,'b) term
  | LetProd of tevar list * ('a, 'b) term * ('a, 'b) term
  | Variant of (('a,'b) typ list) * Datatype.label_id * ('a,'b) term
  | Match of ('a,'b) typ * ('a,'b) term * ('a,'b) branch list

and ('a,'b) branch = ('a,'b) pattern * ('a,'b) term

and ('a,'b) pattern =
  | PVar of tevar
  | PWildcard
  | PInj of (('a,'b) typ list) * int * ('a,'b) pattern
  | PTuple of ('a,'b) pattern list
  | PVariant of (('a,'b) typ list) * Datatype.label_id * ('a,'b) pattern

type nominal_term = (tyvar, tyvar) term
type nominal_pattern = (tyvar, tyvar) pattern
type nominal_branch = nominal_pattern * nominal_term

type debruijn_term =
    (DeBruijn.index, unit) term

(* -------------------------------------------------------------------------- *)

(* Constructors. *)

let ftyabs vs t =
  List.fold_right (fun v t -> TyAbs (v, t)) vs t
let ftyapp t tys =
  List.fold_left (fun t ty -> TyApp (t, ty)) t tys

(* -------------------------------------------------------------------------- *)

(* Boilerplate code for type-in-type substitutions. *)

module TypeVar : DeBruijn.VAR
  with type ('a, 'b) v = ('a, 'b) typ
= struct

  type ('v, 'b) v =
      ('v, 'b) typ

  let var x =
    TyVar x

end

module TypeInType : DeBruijn.TRAVERSE
  with type ('v, 'b) v = ('v, 'b) typ
   and type ('v, 'b) t = ('v, 'b) typ
= struct

  type ('v, 'b) v =
      ('v, 'b) typ

  type ('v, 'b) t =
      ('v, 'b) typ

  (* I note that we could perform physical equality tests in [traverse]
     so as to preserve sharing when possible, but that would require use
     of [Obj.magic]. Let's just not do it in this demo. *)

  let rec traverse lookup extend env ty =
    let trav env = traverse lookup extend env in
    match ty with
    | TyVar x ->
        lookup env x
    | TyArrow (ty1, ty2) ->
        let ty1' = trav env ty1
        and ty2' = trav env ty2 in
        TyArrow (ty1', ty2')
    | TyProduct tys ->
        let tys' = List.map (trav env) tys in
        TyProduct tys'
    | TySum tys ->
        let tys' = List.map (trav env) tys in
        TySum tys'
    | TyConstr (tyconstr, tys) ->
        let tys' = List.map (trav env) tys in
        TyConstr (tyconstr, tys')
    | TyForall (x, ty1) ->
        let env, x = extend env x in
        let ty1' = trav env ty1 in
        TyForall (x, ty1')
    | TyMu (x, ty1) ->
        let env, x = extend env x in
        let ty1' = trav env ty1 in
        TyMu (x, ty1')

end

(* -------------------------------------------------------------------------- *)

(* Boilerplate code for type-in-term substitutions. *)

module TypeInTerm : DeBruijn.TRAVERSE
  with type ('v, 'b) v = ('v, 'b) typ
  with type ('v, 'b) t = ('v, 'b) term
= struct

  type ('v, 'b) v =
      ('v, 'b) typ

  type ('v, 'b) t =
      ('v, 'b) term

  let rec traverse lookup extend env t =
    let trav env = traverse lookup extend env in
    match t with
    | Var x ->
        Var x
    | Abs (x, ty, t) ->
        let ty' = TypeInType.traverse lookup extend env ty
        and t' = trav env t in
        Abs (x, ty', t')
    | App (t1, t2) ->
        let t1' = trav env t1
        and t2' = trav env t2 in
        App (t1', t2')
    | Let (x, t1, t2) ->
        let t1' = trav env t1
        and t2' = trav env t2 in
        Let (x, t1', t2')
    | TyAbs (x, t) ->
        let env, x = extend env x in
        let t' = trav env t in
        TyAbs (x, t')
    | TyApp (t, ty) ->
        let t' = trav env t
        and ty' = TypeInType.traverse lookup extend env ty in
        TyApp (t', ty')
    | Tuple ts ->
        let ts' = List.map (trav env) ts in
        Tuple ts'
    | Proj (i, t) ->
        let t' = trav env t in
        Proj (i, t')
    | Inj (tys, i, t) ->
        let tys' = List.map (TypeInType.traverse lookup extend env) tys
        and t' = trav env t in
        Inj (tys', i, t')
    | LetProd (xs, t1, t2) ->
        let t1' = trav env t1
        and t2' = trav env t2 in
        LetProd (xs, t1', t2')
    | Variant (tys, lbl, t) ->
        let tys' = List.map (TypeInType.traverse lookup extend env) tys
        and t' = trav env t in
        Variant (tys', lbl, t')
    | Match (ty, t, brs) ->
        let ty' = TypeInType.traverse lookup extend env ty in
        let t' = trav env t
        and brs' = List.map (traverse_branch lookup extend env) brs in
        Match (ty', t', brs')

  and traverse_branch lookup extend env (pat,t) =
    (traverse_pattern lookup extend env pat,
     traverse lookup extend env t)

  and traverse_pattern lookup extend env pat =
    let trav env = traverse_pattern lookup extend env in
    match pat with
    | PVar x ->
        PVar x
    | PWildcard ->
        PWildcard
    | PInj (tys, i, pat) ->
        let tys' = List.map (TypeInType.traverse lookup extend env) tys
        and pat' = trav env pat in
        PInj (tys', i, pat')
    | PTuple ps ->
        let ps' = List.map (trav env) ps in
        PTuple ps'
    | PVariant (tys, lbl, pat) ->
        let tys' = List.map (TypeInType.traverse lookup extend env) tys
        and pat' = trav env pat in
        PVariant (tys', lbl, pat')

end

(* -------------------------------------------------------------------------- *)

(* Type-in-type weakening and substitution. *)

module Type = struct
  include DeBruijn.MakeLift(TypeVar)(TypeInType)

  include DeBruijn.MakeSubst(TypeVar)(TypeInType)(TypeInType)

  include DeBruijn.MakeTranslate(TypeVar)(TypeInType)(Int)
end

(* -------------------------------------------------------------------------- *)

(* Translation of nominal terms to de Bruijn terms. *)

module Term = struct
  include DeBruijn.MakeTranslate(TypeVar)(TypeInTerm)(Int)
end


module N2DB = DeBruijn.Nominal2deBruijn(Int)

(* Translate a nominal-F datatype description into a De Bruijn version *)
let translate_datatype tdescr =

  (* We create a De Bruijn environemnt from the nominal type-parameters *)
  let params_env : N2DB.env =
    List.fold_left N2DB.extend N2DB.empty tdescr.Datatype.type_params
  in

  let translate_label ldescr =
    let new_arg_type =
      Type.translate_open params_env ldescr.Datatype.arg_type
    in
    Datatype.{
      label_name = ldescr.label_name;
      type_name = ldescr.type_name;
      arg_type = new_arg_type;
    }
  in

  let new_type_params =
    List.map (fun _param -> ()) tdescr.Datatype.type_params
  in
  let new_labels_descr =
    List.map translate_label tdescr.Datatype.labels_descr
  in

  (* The name and kind of data stay the same, only the type parameters and
     labels description need to change. *)
  Datatype.{
      name = tdescr.name;
      type_params = new_type_params;
      data_kind = tdescr.data_kind;
      labels_descr = new_labels_descr;
  }

let translate_env (env : nominal_datatype_env)
    : debruijn_datatype_env =
  Env.map translate_datatype env
