
(* -------------------------------------------------------------------------- *)

(* Translation from ML types to F types *)

let fresh_tyvar =
  let cpt = ref 0 in

  (* We generate odd number for the user and even number for the solver *)
  fun () -> incr cpt; 2 * !cpt + 1

(* Takes a ML type constructor such as "('a,'b) t",
   a System F type for each parameter ("foo", "bar"),
   and returns a System F type for the translation
   of the type constructor with those parameters
   (morally, "(foo, bar) translation-of-t").

   For example, for the ML declaration
     "type 'a list = Nil of () | Cons of ('a * 'a list)"
   calling translate_type_expr with parameter "int"
   gives the translation
     "mu tau. [{} + {int * tau}]"
*)
let rec translate_type_expr
  (env : ML.datatype_env)
  (params : F.nominal_type list)
  (type_id : Datatype.tyconstr_id)
  : F.nominal_type =
  let desc = Env.type_descr_from_type env type_id in
  let tau = fresh_tyvar () in
  let rec_occur = F.TyVar tau in
  let tys = translate_typedecl env params desc rec_occur in
  F.TyMu (tau, F.TySum tys)

(* Takes a ML type constructor such as "('a,'b) t",
   a System F type for each parameter ("foo", "bar"),
   and returns a list of System F types for the arguments
   of the constructors of "(foo,bar) t".

   For example, for the ML declaration
     "type 'a list = Nil of () | Cons of ('a * 'a list)"
   calling translate_type_expr with parameter "int"
   gives a list containing the constructor types
     "{}"
   and
     "{int * mu tau. [{} + {int * tau}]}".
*)
and translate_constr_types
  (env : ML.datatype_env)
  (params : F.nominal_type list)
  (type_id : Datatype.tyconstr_id)
  : F.nominal_type list =
  let desc = Env.type_descr_from_type env type_id in
  let rec_occur = translate_type_expr env params type_id in
  translate_typedecl env params desc rec_occur

(* Here [rec_occur] is the value provide when we encounter an
   occurence of the name of the type we translate *)
and translate_type
  (env : ML.datatype_env)
  (params : (string * F.nominal_type) list)
  (rec_occur : F.nominal_type)
  (name_self : Datatype.tyconstr_id)
  (ty : ML.typ)
  : F.nominal_type =
  let rec tt (ty : ML.typ) : F.nominal_type =
    match ty with
    | ML.TyVar x ->
        List.assoc x params

    | ML.TyArrow (ty1, ty2) ->
        F.TyArrow (tt ty1, tt ty2)

    | ML.TyProduct tys ->
        F.TyProduct (List.map tt tys)

    | ML.TyConstr (tid, tys) ->
        if tid = name_self then begin
            let desc = Env.type_descr_from_type env name_self in

            (* This assert should hold as long as we only treat
             * regular recursion, that is types in which each type parameters
             * is instantiated with only one type *)
            assert (tys = List.map (fun x -> ML.TyVar x) desc.Datatype.type_params);
            rec_occur
          end
        else
          translate_type_expr env (List.map tt tys) tid
  in
  tt ty

and translate_typedecl
  (env : ML.datatype_env)
  (params : F.nominal_type list)
  (desc : (string, ML.typ) Datatype.datatype_descr)
  (rec_occur : F.nominal_type)
    : F.nominal_type list =
  let params = List.combine Datatype.(desc.type_params) params in
  let tys =
    List.map (fun ldescr -> Datatype.(ldescr.arg_type)) Datatype.(desc.labels_descr)
    |> List.map (translate_type env params rec_occur Datatype.(desc.name))
  in
  tys

(* Translation of ML datatype environment into System F datatype environment. *)

(* Given a way to translate ML type variables into System F type variables,
   this function translates an ML type into a System F type. *)
let rec translate_type (translate_tyvar : ML.tyvar -> F.tyvar) (ty : ML.typ)
    : F.nominal_type =
  let trans ty = translate_type translate_tyvar ty in
  match ty with
  | ML.TyVar tv ->
      F.TyVar (translate_tyvar tv)
  | ML.TyArrow (ty1, ty2) ->
      F.TyArrow (trans ty1, trans ty2)
  | ML.TyProduct tys ->
      F.TyProduct (List.map trans tys)
  | ML.TyConstr (tyconstr, tys) ->
      F.TyConstr (tyconstr, List.map trans tys)

(* Translate a ML datatype description into a nominal System F version *)
let translate_datatype tdescr =

  (* We create an association list between ML and System F type variables.
     We'll use this association list to translate type parameters and
     type arguments containing those type parameters. *)
  let assoc =
    List.map (fun tyvar -> (tyvar, fresh_tyvar ()))
      tdescr.Datatype.type_params
  in

  let translate_label ldescr =
    let new_arg_type =
      translate_type (fun tv -> List.assoc tv assoc) ldescr.Datatype.arg_type
    in
    Datatype.{
      label_name = ldescr.label_name;
      type_name = ldescr.type_name;
      arg_type = new_arg_type;
    }
  in

  let new_type_params =
    List.map (fun param -> List.assoc param assoc) tdescr.Datatype.type_params
  in
  let new_labels_descr =
    List.map translate_label tdescr.Datatype.labels_descr
  in

  (* The name and kind of data stay the same, only the type parameters and
     labels description need to change. *)
  Datatype.{
      name = tdescr.name;
      type_params = new_type_params;
      data_kind = tdescr.data_kind;
      labels_descr = new_labels_descr;
  }

let translate_env (env : ML.datatype_env)
    : F.nominal_datatype_env =
  Env.map translate_datatype env
